"""
JSON Encoder Utility
"""

import json
import datetime


class JSONEncoder(json.JSONEncoder):
    """
    Custom JSON encoder to handle objects like datetime before they are sent to the API.
    """

    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)  # pragma: no cover



"""
Data model for supported services
"""

services = [
    "accounts",
    "campaigns",
    "contacts",
    "creatives",
    "custom_fields",
    "external_users",
    "groups",
    "line_items",
    "line_item_creative_associations",
    "opportunities",
    "opportunity_line_items",
    "packages",
    "products",
    "product_rates",
    "rate_cards",
    "reports",
    "users",
]



"""
Models for the PlacementsIO_OAuth
"""

from typing import List, Literal
from pio.model.service import services

service_scopes = [f"{service}_read" for service in services] + [
    f"{service}_write" for service in services
]
ModelScopes = List[Literal[tuple(service_scopes)]]



"""
Placements.io Python SDK
"""

import os
import logging
import datetime
import csv
import time
from typing import Unpack, Union
from pio.client import PlacementsIOClient
from pio.error.api_error import APIError
from pio.model.environment import API
from pio.model.report import COLUMNS
from pio.model.get import (
    ModelFilterDefaults,
    ModelFilterAccount,
    ModelFilterCampaign,
    ModelFilterContact,
    ModelFilterCreative,
    ModelFilterCustomField,
    ModelFilterGroup,
    ModelFilterLineItem,
    ModelFilterOpportunity,
    ModelFilterOpportunityLineItem,
    ModelFilterPackage,
    ModelFilterProduct,
    ModelFilterProductRate,
    ModelFilterRateCard,
    ModelFilterReport,
    ModelFilterUser,
)
import httpx


class PlacementsIO:
    """
    Placements.io Python SDK
    """

    def __init__(self, environment: str = None, token: str = None):
        environment = (
            environment or os.environ.get(f"PLACEMENTS_IO_ENVIRONMENT") or "staging"
        )
        self.base_url = API.get(environment, environment)
        self.token = (
            token
            or os.environ.get(f"PLACEMENTS_IO_TOKEN_{environment.upper()}")
            or os.environ.get("PLACEMENTS_IO_TOKEN")
        )
        self.logger = logging.getLogger("pio")
        self.settings = {
            "base_url": self.base_url,
            "token": self.token,
        }

    def relationship(self, relationship_url: str):
        """
        Returns a Service class from a relationship URL provided for a previous API call
        """
        return self.Service(
            **{
                "base_url": self.base_url,
                "token": self.token,
                "service": relationship_url.replace(self.base_url, ""),
                "model": {"get": ModelFilterDefaults},
            }
        )

    class Service(PlacementsIOClient):
        """
        Class for interacting with API Services
        """

        def __init__(self, token, base_url, service, model, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.token = token
            self.base_url = base_url
            self.service = service
            self.model = model

        async def get(
            self,
            include: list = None,
            fields: list = None,
            params: dict = None,
            **args: Unpack[ModelFilterAccount],
        ) -> list:
            """
            Get existing resources within the service
            """
            return await self.client(
                service=self.service,
                includes=include,
                filters=args,
                fields=fields,
                param=params,
            )

        async def update(
            self,
            resource_ids: list,
            attributes: Union[callable, dict] = None,
            relationships: Union[callable, dict] = None,
            params: dict = None,
        ) -> dict:
            """
            Update existing resources within the service
            """
            return await self.client_update(
                service=self.service,
                resource_ids=resource_ids,
                attributes=attributes,
                relationships=relationships,
                params=params,
            )

        async def create(
            self,
            objects: list[dict],
        ) -> dict:
            """
            Create new resources within the service
            """
            return await self.client_create(
                service=self.service,
                objects=objects,
            )

    class ReportService(Service):
        """
        Class for interacting with API Reports Service
        """

        TODAY_START = datetime.datetime.now(datetime.timezone.utc).replace(
            hour=0, minute=0, second=0, microsecond=0
        )
        TODAY_END = datetime.datetime.now(datetime.timezone.utc).replace(
            hour=23, minute=59, second=59, microsecond=999
        )

        async def create(
            self, start_date=TODAY_START, end_date=TODAY_END, columns: list = COLUMNS
        ) -> int:
            """
            Create a new report
            """
            report_creation_request = [
                {
                    "attributes": {
                        "definition": {
                            "start-date": start_date,
                            "end-date": end_date,
                            "columns": columns,
                        }
                    }
                }
            ]
            report_creation = await self.client_create(
                service=self.service, objects=report_creation_request
            )
            report_id = report_creation[0].get("id")
            return report_id

        async def get(self, resource_id: int) -> dict:
            """
            Get existing resources within the service
            """
            return await self.resource(service=self.service, resource_id=resource_id)

        async def data(self, report_id: dict) -> list:
            """
            Returns report data in a list of dictionaries
            """

            report_response = await self.get(report_id)
            status = report_response.get("attributes", {}).get("status")
            while status in ["pending", "in_progress"]:
                self.logger.info(f"Report is currently {status}. Retrying in 5 seconds")
                time.sleep(5)
                report_response = await self.get(report_id)
                status = report_response.get("attributes", {}).get("status")
            if status == "failed":
                raise APIError(
                    report_response.get("attributes", {}).get("error-message")
                )

            download_url = report_response.get("attributes", {}).get("download-url")
            if not download_url:
                raise APIError(
                    "No download report URL found in response. Unable to download report data.",
                    report_response,
                )
            async with httpx.AsyncClient() as data_client:
                async with data_client.stream(
                    "GET", download_url, follow_redirects=True
                ) as response:
                    response.raise_for_status()
                    lines = []
                    async for line in response.aiter_lines():
                        lines.append(line)
                    csv_reader = csv.reader(lines)
                    headers = next(csv_reader)
                    rows = []
                    for row in csv_reader:
                        rows.append(dict(zip(headers, row)))
                    return rows

    async def oauth2(self, client_id: str, redirect_url: str) -> Service:
        """
        Returns an OAuth2 Service object for use in interacting with the API
        """
        return await self.Service(
            **self.settings, service="oauth/authorize", model={}
        ).create(
            [
                {
                    "client_id": client_id,
                    "redirect_url": redirect_url,
                    "response_type": "code",
                }
            ]
        )

    @property
    def accounts(self) -> Service:
        """
        Returns an Accounts Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="accounts", model={"get": ModelFilterAccount}
        )

    @property
    def campaigns(self) -> Service:
        """
        Returns a Campaigns Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="campaigns", model={"get": ModelFilterCampaign}
        )

    @property
    def contacts(self) -> Service:
        """
        Returns a Contacts Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="contacts", model={"get": ModelFilterContact}
        )

    @property
    def creatives(self) -> Service:
        """
        Returns a Creatives Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="creatives", model={"get": ModelFilterCreative}
        )

    @property
    def custom_fields(self) -> Service:
        """
        Returns a Custom Fields Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="custom_fields",
            model={"get": ModelFilterCustomField},
        )

    @property
    def external_users(self) -> Service:
        """
        Returns a External Users Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="external_users",
            model={"get": ModelFilterCustomField},
        )

    @property
    def groups(self) -> Service:
        """
        Returns a Groups Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="groups", model={"get": ModelFilterGroup}
        )

    @property
    def line_items(self) -> Service:
        """
        Returns a Line Items Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="line_items", model={"get": ModelFilterLineItem}
        )

    @property
    def line_item_creative_associations(self) -> Service:
        """
        Returns a Line Item Creative Associations Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="line_item_creative_associations",
            model={"get": ModelFilterDefaults},
        )

    @property
    def opportunities(self) -> Service:
        """
        Returns an Opportunities Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="opportunities",
            model={"get": ModelFilterOpportunity},
        )

    @property
    def opportunity_line_items(self) -> Service:
        """
        Returns an Opportunity Line Items Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="opportunity_line_items",
            model={"get": ModelFilterOpportunityLineItem},
        )

    @property
    def packages(self) -> Service:
        """
        Returns a Packages Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="packages", model={"get": ModelFilterPackage}
        )

    @property
    def products(self) -> Service:
        """
        Returns a Products Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="products", model={"get": ModelFilterProduct}
        )

    @property
    def product_rates(self) -> Service:
        """
        Returns a Product Rates Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="product_rates",
            model={"get": ModelFilterProductRate},
        )

    @property
    def rate_cards(self) -> Service:
        """
        Returns a Rate Cards Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings,
            service="rate_cards",
            model={"get": ModelFilterRateCard},
        )

    @property
    def reports(self) -> Service:
        """
        Returns a Reports Service object for use in interacting with the API
        """
        return self.ReportService(
            **self.settings, service="reports", model={"get": ModelFilterReport}
        )

    @property
    def users(self) -> Service:
        """
        Returns a Users Service object for use in interacting with the API
        """
        return self.Service(
            **self.settings, service="users", model={"get": ModelFilterUser}
        )



"""
Placements.io Python SDK
Low level code to interact with the Placements.io Python SDK
"""

import logging
import asyncio
import json
from typing import Union
import httpx
from pio.error.api_error import APIError
from pio.utility.json_encoder import JSONEncoder


class PlacementsIOClient:
    """
    Placements.io Python SDK
    Low level code to interact with the Placements.io API
    """

    def __init__(self):
        self.logger = logging.getLogger("pio")
        self.base_url = None
        self.token = None

    @property
    def _version(self):
        """
        Returns the version of the client library
        Note: This is locally imported to avoid circular imports
        """
        from pio import __version__  # pylint: disable=import-outside-toplevel

        return __version__

    def pagination(self, page_number: int = 1) -> dict:
        """
        Provides pagination parameters for the API request.
        """
        return {
            "page[number]": page_number,
            "page[size]": 100,
        }

    def headers(self, method, service, is_retry) -> dict:
        """
        Returns standardized headers for the API request.
        """
        token = self.token
        if callable(self.token):
            token = self.token()
        return {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/vnd.api+json",
            "User-Agent": f"PlacementsIO Python SDK/{self._version}",
            "x-metadata": json.dumps(
                {
                    "method": method,
                    "service": service.split("/")[0],
                    "is_retry": is_retry,
                }
            ),
        }

    async def client_request(
        self,
        client: httpx.AsyncClient,
        method: str,
        resource: str,
        request: dict,
        is_retry: bool = False,
    ) -> httpx.Response:
        client_method = getattr(client, method)
        request = {
            "url": resource,
            "headers": self.headers(method, resource, is_retry),
            **request,
        }
        if request.get("data") and not isinstance(request["data"], str):
            request["data"] = json.dumps(request["data"], default=str, cls=JSONEncoder)
        response = await client_method(**request)
        if response.status_code == 429:
            retry_after = int(response.headers.get("Retry-After", 60))
            self.logger.warning(
                "Rate limit reached. Waiting %s seconds before retrying...",
                retry_after,
            )
            await asyncio.sleep(retry_after)
            return await self.client_request(
                client, method, resource, request, is_retry=True
            )
        return response

    async def client(
        self,
        service: str,
        param: dict = None,
        filters: dict = None,
        includes: list = None,
        fields: list = None,
    ) -> list:
        """
        Get existing resources within the service
        """
        # TODO: Need to have a way to call multiple IDs at the same time
        async with httpx.AsyncClient(base_url=self.base_url, timeout=60) as client:

            param = param or {}
            param.update(self.pagination())
            param.update(self._filter_values(filters))
            param.update(self._list_values("include", includes))
            param.update(
                self._list_values(f"fields[{service.replace('_', '-')}]", fields)
            )
            self.logger.info("Fetching data from %s", service)
            response = await self.client_request(
                client, "get", service, {"params": param}
            )
            data = response.json()
            errors = data.get("errors", [])
            if errors:
                raise APIError(errors)
            results = data.get("data", [])
            meta = data.get("meta", {})
            page_count = meta.get("page-count", 0)
            if page_count > 1:
                self.logger.info(
                    "Paginating data from %s [%s Pages]", service, page_count
                )

            tasks = []
            for page_number in range(2, page_count + 1):
                paginated_param = param.copy()
                paginated_param.update(self.pagination(page_number))
                tasks.append(
                    self.client_request(
                        client, "get", service, {"params": paginated_param}
                    )
                )
            responses = await asyncio.gather(*tasks)
            for response in responses:
                data = response.json()
                data = data.get("data", [])
                results.extend(data)

            return results

    async def resource(
        self,
        service: str,
        resource_id: int,
    ) -> dict:
        """
        Get a single existing resource within the service
        """

        # Follow_redirects is set to True to facilitate report downloads
        async with httpx.AsyncClient(
            base_url=self.base_url, timeout=60, follow_redirects=True
        ) as client:
            path = f"{service}/{resource_id}"
            self.logger.info("Fetching data from %s %s", service, resource_id)
            response = await self.client_request(client, "get", path, {})
            data = response.json()
            errors = data.get("errors", [])
            if errors:
                raise APIError(errors)
            results = data.get("data", {})
            return results

    async def client_update(
        self,
        service: str,
        resource_ids: list,
        attributes: Union[callable, dict] = None,
        relationships: Union[callable, dict] = None,
        params: dict = None,
    ) -> dict:
        """
        Update existing resources within the service
        """
        if not attributes and not relationships:
            raise ValueError(
                "Must provide either attributes or relationships to update."
            )

        async def get_responses(resource_ids: list) -> dict:

            async def make_multiple_requests(resource_ids: int) -> dict:

                async with httpx.AsyncClient(
                    base_url=self.base_url, timeout=60
                ) as client:
                    tasks = []
                    for resource_id in resource_ids:
                        url = f"{service}/{resource_id}"

                        attributes_payload = {}
                        if isinstance(attributes, dict):
                            attributes_payload = {"attributes": attributes}
                        elif callable(attributes):
                            attributes_payload = {
                                "attributes": await attributes(resource_id)
                            }

                        relationships_payload = {}
                        if isinstance(relationships, dict):
                            relationships_payload = {"relationships": relationships}
                        elif callable(relationships):
                            relationships_payload = {
                                "relationships": await relationships(resource_id)
                            }

                        payload = {
                            "data": {
                                "id": resource_id,
                                "type": service,
                                **attributes_payload,
                                **relationships_payload,
                            }
                        }
                        self.logger.info(
                            "Updating %s %s",
                            url,
                            params,
                        )
                        self.logger.debug(
                            "Payload: %s",
                            json.dumps(payload, indent=4, default=str, cls=JSONEncoder),
                        )
                        tasks.append(
                            self.client_request(
                                client,
                                "patch",
                                url,
                                {"data": payload, "params": params},
                            )
                        )
                    return await asyncio.gather(*tasks)

            responses_dict = {}
            responses = await make_multiple_requests(resource_ids)
            responses_dict.update(dict(zip(resource_ids, responses)))
            return responses_dict

        # Split resource_ids into chunks of 100
        chunk_size = 100
        raw_responses = {}
        if isinstance(resource_ids, {}.keys().__class__):
            resource_ids = list(resource_ids)
        for index in range(0, len(resource_ids), chunk_size):
            chunk = resource_ids[index : index + chunk_size]
            chunk_responses = await get_responses(chunk)
            raw_responses.update(chunk_responses)

        expanded_responses = [
            (
                response.json().get(
                    "data", {**response.json(), "links": {"self": response.request.url}}
                )
                if response.content
                else {
                    "errors": [
                        {
                            "title": "No data",
                            "detail": "No data was returned in the API response",
                        }
                    ],
                    "links": {"self": response.request.url},
                }
            )
            for response in raw_responses.values()
        ]
        return expanded_responses

    async def client_create(
        self,
        service: str,
        objects: list[dict],
    ) -> dict:
        """
        Create new resources within the service
        """

        async def get_responses(objects: list) -> list:

            async def make_multiple_requests(objects) -> list:
                async with httpx.AsyncClient(
                    base_url=self.base_url, timeout=60
                ) as client:
                    tasks = []
                    for resources in objects:
                        attributes = resources.get("attributes")
                        relationships = resources.get("relationships")

                        attributes_payload = {}
                        if isinstance(attributes, dict):
                            attributes_payload = {"attributes": attributes}

                        relationships_payload = {}
                        if isinstance(relationships, dict):
                            relationships_payload = {"relationships": relationships}

                        payload = {
                            "data": {
                                "type": service,
                                **attributes_payload,
                                **relationships_payload,
                            }
                        }
                        tasks.append(
                            self.client_request(
                                client, "post", service, {"data": payload}
                            )
                        )
                    return await asyncio.gather(*tasks)

            attempts = 0
            retry_after = 60
            responses_final = []
            while len(objects) != len(responses_final):
                self.logger.info(objects)
                if attempts:
                    self.logger.warning("%s Resources remaining...", len(objects))
                    self.logger.warning(
                        "Waiting %s seconds before retrying...", retry_after
                    )
                    await asyncio.sleep(retry_after)
                responses = await make_multiple_requests(objects)

                # Now look for 429 responses to retry, or update the response with JSON
                for response in responses:
                    if response.status_code != 429:
                        responses_final.append(response)
                attempts += 1
            return responses_final

        raw_responses = await get_responses(objects)
        expanded_responses = [
            response.json().get("data", response.json()) for response in raw_responses
        ]
        return expanded_responses

    def _filter_values(self, params: dict = None) -> str:
        params = params or {}
        return {f"filter[{key}]": value for key, value in params.items()}

    def _list_values(self, key: str, relationships: list = None) -> dict:
        params = {}
        if relationships:
            params = {key: ",".join(relationships)}
        return params



"""
Data model for the report service columns
"""

COLUMNS = [
    "ad_server",
    "ad_server_id",
    "first_party_impressions",
    "first_party_clicks",
    "first_party_ctr",
    "first_party_delivered_amount",
    "third_party_impressions",
    "third_party_clicks",
    "third_party_ctr",
    "third_party_delivered_amount",
    "reconciled_units",
    "reconciled_amount",
    "billed_units",
    "billed_amount",
]



"""
Models for the GET API Endpoints
"""

from typing import TypedDict
from datetime import datetime


class ModelFilterDefaults(TypedDict):
    """
    Default API Filters that exist for all services
    """

    id: int
    modified_since: datetime
    name: str
    uid: str


class ModelFilterAccount(ModelFilterDefaults):
    """
    API Filters for Accounts
    """

    account_type: str
    external_id: str
    archived: bool
    advertiser_of_record: int
    agency_of_record: int


class ModelFilterCampaign(ModelFilterDefaults):
    """
    API Filters for Campaigns
    """

    ad_server_network_code: int
    ad_server_id: int
    archived: bool
    campaign_number: int


class ModelFilterContact(ModelFilterDefaults):
    """
    API Filters for Contacts
    """


class ModelFilterCreative(ModelFilterDefaults):
    """
    API Filters for Creatives
    """


class ModelFilterCustomField(ModelFilterDefaults):
    """
    API Filters for Custom Fields
    """


class ModelFilterGroup(ModelFilterDefaults):
    """
    API Filters for Groups
    """

    ad_server_network_code: int
    ad_server_id: int
    campagign: int


class ModelFilterLineItem(ModelFilterDefaults):
    """
    API Filters for Line Items
    """

    ad_server_network_code: int
    archived: bool
    approval_status: str
    delivery_status: str
    started_before: datetime
    started_after: datetime
    ended_before: datetime
    ended_after: datetime
    ad_server_id: int
    campaign: int
    group: int


class ModelFilterOpportunity(ModelFilterDefaults):
    """
    API Filters for Opportunities
    """

    archived: bool
    opportunity_order_number: int


class ModelFilterOpportunityLineItem(ModelFilterDefaults):
    """
    API Filters for Opportunity Line Items
    """

    ad_server_network_code: int
    archived: bool
    started_before: datetime
    started_after: datetime
    ended_before: datetime
    ended_after: datetime
    opportunity: int


class ModelFilterPackage(ModelFilterDefaults):
    """
    API Filters for Packages
    """

    active: bool
    archived: bool


class ModelFilterProduct(ModelFilterDefaults):
    """
    API Filters for Products
    """

    ad_server: str
    ad_server_network_code: int
    active: bool
    archived: bool


class ModelFilterProductRate(ModelFilterDefaults):
    """
    API Filters for Product Rates
    """


class ModelFilterRateCard(ModelFilterDefaults):
    """
    API Filters for Rate Cards
    """


class ModelFilterReport(ModelFilterDefaults):
    """
    API Filters for Reports
    """


class ModelFilterUser(ModelFilterDefaults):
    """
    API Filters for Users
    """

    email: str



import http.server
import socketserver
import urllib.parse
import webbrowser
import logging
import datetime
import httpx
import time
from pio.model.environment import API
from pio.model.oauth import ModelScopes
from pio.pio import PlacementsIO

OAUTH_RESPONSE = None


class OAuthMiniServer(http.server.SimpleHTTPRequestHandler):  # pragma: no cover
    def do_GET(self):
        # Parse the query parameters
        parsed_path = urllib.parse.urlparse(self.path)
        query_params = urllib.parse.parse_qs(parsed_path.query)

        # Extract the token code from the query parameters
        global OAUTH_RESPONSE  # pylint: disable=global-statement
        OAUTH_RESPONSE = query_params.get("code", [None])[0]
        logging.debug("Received OAuth token code", OAUTH_RESPONSE)

        # Respond to the client
        if OAUTH_RESPONSE:
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"OAuth redirect received! You may now close this window.")
        else:
            error = query_params.get("error", ["OAuth authentication failed."])[0]
            error_description = query_params.get(
                "error_description", ["Please try again."]
            )[0]
            self.send_response(401)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(
                f"""
                {error}<br>
                {error_description}<br>
                <br>
                <a href='javascript:history.go(-1)'>Try again</a>
                """.encode()
            )


class PlacementsIO_OAuth(PlacementsIO):
    def __init__(
        self,
        environment: str = None,
        application_id: str = None,
        client_secret: str = None,
        redirect_host: str = "http://localhost",
        redirect_port: int = 17927,
        scopes: ModelScopes = None,
    ):
        self.base_url = API[environment]
        self.oauth_base_url = self.base_url.replace("/v1/", "/oauth/")
        self.application_id = application_id
        self.client_secret = client_secret
        self.redirect_host = redirect_host
        self.redirect_port = redirect_port
        self._oauth = {
            "access_token": None,
            "expires_in": 0,
            "refresh_token": None,
        }
        self._oauth_expiry = None
        self._set_oauth_expiry_time()
        scopes_param = f"&scope={'+'.join(scopes or [])}" if scopes else ""
        self.auth_url = (
            f"{self.oauth_base_url}authorize"
            f"?client_id={self.application_id}"
            f"&redirect_uri={self.redirect_host}:{self.redirect_port}"
            f"&response_type=code{scopes_param}"
        )
        self.settings = {
            "base_url": self.base_url,
            "token": self.token,
        }
        self.logger = logging.getLogger("pio")

    def get_auth_url(self):
        return self.auth_url

    def get_user_auth(self):
        with socketserver.TCPServer(("", self.redirect_port), OAuthMiniServer) as httpd:
            print("Opening browser to authenticate...")
            print(f"\t{self.auth_url}")
            webbrowser.open(self.auth_url)
            print(
                f"Waiting for OAuth response on {self.redirect_host}:{self.redirect_port}"
            )
            print("\tPress Ctrl+C to cancel.")
            httpd.handle_request()
            while OAUTH_RESPONSE is None:
                time.sleep(1)
        return self.set_user_auth(OAUTH_RESPONSE)

    def set_user_auth(self, auth_code):
        self._fetch_access_token(auth_code)

    def token(self):
        if (
            datetime.datetime.now() > self._oauth_expiry
            and self._oauth.get("refresh_token") is not None
        ):
            self.logger.debug("Refreshing access token with refresh token")
            self._fetch_access_token(
                self._oauth.get("refresh_token"), grant="refresh_token"
            )
        elif self._oauth.get("access_token") is None:
            self.get_user_auth()
        return self._oauth.get("access_token")

    def _set_oauth_expiry_time(self, seconds=0):
        self._oauth_expiry = datetime.datetime.now() + datetime.timedelta(
            seconds=seconds
        )

    def _fetch_access_token(self, auth_code, grant: str = "authorization_code"):
        field_name = "code" if grant == "authorization_code" else grant
        response = httpx.post(
            url=f"{self.oauth_base_url}token",
            data={
                "client_id": self.application_id,
                "client_secret": self.client_secret,
                "redirect_uri": f"{self.redirect_host}:{self.redirect_port}",
                "grant_type": grant,
                field_name: auth_code,
            },
        )
        self._oauth = response.json()
        self._set_oauth_expiry_time(self._oauth.get("expires_in"))
